import numpy as np
from lensless.io import load_data
from lensless import GradientDescient, NesterovGradientDescent, FISTA, ADMM, APGD


psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"
downsample = 16
n_iter = 5
disp = None


def test_algo():
    for algo in [GradientDescient, NesterovGradientDescent, FISTA, ADMM, APGD]:
        for gray in [True, False]:
            for dtype in [np.float32, np.float64]:
                psf, data = load_data(
                    psf_fp=psf_fp,
                    data_fp=data_fp,
                    downsample=downsample,
                    plot=False,
                    gray=gray,
                    dtype=dtype,
                )
                if algo == APGD:
                    if not gray:
                        continue
                recon = algo(psf, dtype=dtype, realconv=False)
                recon.set_data(data)
                res = recon.apply(n_iter=n_iter, disp_iter=None, plot=False)
                if gray:
                    assert len(psf.shape) == 2
                else:
                    assert len(psf.shape) == 3
                assert res.dtype == dtype, f"Got {res.dtype}, expected {dtype}"


test_algo()



import numpy as np
from pycsou.core.linop import LinearOperator
from typing import Union, Optional
from numbers import Number
from scipy import fft
from scipy.fftpack import next_fast_len


class RealFFTConvolve2D(LinearOperator):
    def __init__(self, filter, dtype: Optional[type] = None):
        """
        Linear operator that performs convolution in Fourier domain, and assumes
        real-valued signals.

        Parameters
        ----------
        filter :py:class:`~numpy.ndarray`
            2D filter to use. Must be of shape (height, width, channels) even if
            only one channel.
        dtype : float32 or float64
            Data type to use for optimization.
        """

        assert len(filter.shape) == 3
        self._filter_shape = np.array(filter.shape)
        self._n_channels = filter.shape[2]

        # cropping / padding indices
        self._padded_shape = 2 * self._filter_shape[:2] - 1
        self._padded_shape = np.array([next_fast_len(i) for i in self._padded_shape])
        self._padded_shape = np.r_[self._padded_shape, [self._n_channels]]
        self._start_idx = (self._padded_shape[:2] - self._filter_shape[:2]) // 2
        self._end_idx = self._start_idx + self._filter_shape[:2]

        # precompute filter in frequency domain
        self._H = fft.rfft2(self._pad(filter), axes=(0, 1))
        self._Hadj = np.conj(self._H)
        self._padded_data = np.zeros(self._padded_shape).astype(dtype)

        shape = (int(np.prod(self._filter_shape)), int(np.prod(self._filter_shape)))
        super(RealFFTConvolve2D, self).__init__(shape=shape, dtype=dtype)

    def _crop(self, x):
        return x[self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]]

    def _pad(self, v):
        vpad = np.zeros(self._padded_shape).astype(v.dtype)
        vpad[self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]] = v
        return vpad

    def __call__(self, x: Union[Number, np.ndarray]) -> Union[Number, np.ndarray]:
        # like here: https://github.com/PyLops/pylops/blob/3e7eb22a62ec60e868ccdd03bc4b54806851cb26/pylops/signalprocessing/ConvolveND.py#L103
        self._padded_data[
            self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]
        ] = np.reshape(x, self._filter_shape)
        y = self._crop(
            fft.ifftshift(
                fft.irfft2(fft.rfft2(self._padded_data, axes=(0, 1)) * self._H, axes=(0, 1)),
                axes=(0, 1),
            )
        )
        return y.ravel()

    def adjoint(self, y: Union[Number, np.ndarray]) -> Union[Number, np.ndarray]:
        self._padded_data[
            self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]
        ] = np.reshape(y, self._filter_shape)
        x = self._crop(
            fft.ifftshift(
                fft.irfft2(fft.rfft2(self._padded_data, axes=(0, 1)) * self._Hadj, axes=(0, 1)),
                axes=(0, 1),
            )
        )
        return x.ravel()



import numpy as np
from lensless.recon import ReconstructionAlgorithm
from scipy import fft


class ADMM(ReconstructionAlgorithm):
    """
    Object for applying ADMM (Alternating Direction Method of Multipliers) with
    a non-negativity constraint and a total variation (TV) prior.

    Paper about ADMM: https://web.stanford.edu/~boyd/papers/pdf/admm_distr_stats.pdf
    Slides about ADMM: https://web.stanford.edu/class/ee364b/lectures/admm_slides.pdf

    """

    def __init__(
        self,
        psf,
        dtype=np.float32,
        mu1=1e-6,
        mu2=1e-5,
        mu3=4e-5,
        tau=0.0001,
        psi=None,
        psi_adj=None,
        psi_gram=None,
        **kwargs
    ):
        """

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray`
            Point spread function (PSF) that models forward propagation.
            2D (grayscale) or 3D (RGB) data can be provided and the shape will
            be used to determine which reconstruction (and allocate the
            appropriate memory).
        dtype : float32 or float64
            Data type to use for optimization.
        mu1 : float
            Step size for updating primal/dual variables.
        mu2 : float
            Step size for updating primal/dual variables.
        mu3 : float
            Step size for updating primal/dual variables.
        tau : float
            Weight for L1 norm of `psi` applied to the image estimate.
        psi : :py:class:`function`, optional
            Operator to map image to a space that the image is assumed to be
            sparse in (hence L1 norm). Default is to use total variation (TV)
            operator.
        psi_adj : :py:class:`function`
            Adjoint of `psi`.
        psi_gram : :py:class:`function`
            Function to compute gram of `psi`.
        """
        self._mu1 = mu1
        self._mu2 = mu2
        self._mu3 = mu3
        self._tau = tau

        # call reset() to initialize matrices
        super(ADMM, self).__init__(psf, dtype)

        # set prior
        if psi is None:
            # use already defined Psi and PsiT
            self._PsiTPsi = finite_diff_gram(self._padded_shape, self._dtype)
        else:
            assert psi_adj is not None
            assert psi_gram is not None
            assert callable(psi)
            assert callable(psi_adj)
            assert callable(psi_gram)
            # overwrite already defined Psi and PsiT
            self._Psi = psi
            self._PsiT = psi_adj
            self._PsiTPsi = psi_gram(self._padded_shape)

        # precompute_R_divmat
        self._R_divmat = 1.0 / (
            self._mu1 * (np.abs(np.conj(self._H) * self._H))
            + self._mu2 * np.abs(self._PsiTPsi)
            + self._mu3
        ).astype(self._complex_dtype)

    def _Psi(self, x):
        """
        Operator to map image to space that the image is assumed to be sparse
        in.
        """
        return finite_diff(x)

    def _PsiT(self, U):
        """
        Adjoint of `_Psi`.
        """
        return finite_diff_adj(U)

    def _crop(self, x):
        return x[self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]]

    def _pad(self, v):
        """adjoint of cropping"""
        vpad = np.zeros(self._padded_shape).astype(v.dtype)
        vpad[self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]] = v
        return vpad

    def _forward(self):
        """Convolution with frequency response."""
        return fft.ifftshift(
            fft.irfft2(
                fft.rfft2(self._image_est, axes=(0, 1), s=self._padded_shape[:2]) * self._H,
                axes=(0, 1),
                s=self._padded_shape[:2],
            ),
            axes=(0, 1),
        )

    def _backward(self, x):
        """adjoint of forward / convolution"""
        return fft.ifftshift(
            fft.irfft2(
                fft.rfft2(x, axes=(0, 1), s=self._padded_shape[:2]) * np.conj(self._H),
                axes=(0, 1),
                s=self._padded_shape[:2],
            ),
            axes=(0, 1),
        )

    def reset(self):
        # spatial frequency response
        self._H = fft.rfft2(self._pad(self._psf), axes=(0, 1), s=self._padded_shape[:2]).astype(
            self._complex_dtype
        )

        self._X = np.zeros(self._padded_shape, dtype=self._dtype)
        # self._U = np.zeros(np.r_[self._padded_shape, [2]], dtype=self._dtype)
        self._image_est = np.zeros_like(self._X)
        self._U = np.zeros_like(self._Psi(self._image_est))
        self._W = np.zeros_like(self._X)
        if self._image_est.max():
            # if non-zero
            self._forward_out = self._forward()
            self._Psi_out = self._Psi(self._image_est)
        else:
            self._forward_out = np.zeros_like(self._X)
            self._Psi_out = np.zeros_like(self._U)

        self._xi = np.zeros_like(self._image_est)
        self._eta = np.zeros_like(self._U)
        self._rho = np.zeros_like(self._X)

        # precompute_X_divmat
        self._X_divmat = 1.0 / (self._pad(np.ones(self._psf_shape, dtype=self._dtype)) + self._mu1)

    def _U_update(self):
        """Total variation update."""
        # to avoid computing sparse operator twice
        self._U = soft_thresh(self._Psi_out + self._eta / self._mu2, self._tau / self._mu2)

    def _X_update(self):
        # to avoid computing forward model twice
        self._X = self._X_divmat * (self._xi + self._mu1 * self._forward_out + self._data)

    def _image_update(self):
        rk = (
            (self._mu3 * self._W - self._rho)
            + self._PsiT(self._mu2 * self._U - self._eta)
            + self._backward(self._mu1 * self._X - self._xi)
        )
        freq_space_result = self._R_divmat * fft.rfft2(rk, axes=(0, 1), s=self._padded_shape[:2])
        self._image_est = fft.irfft2(freq_space_result, axes=(0, 1), s=self._padded_shape[:2])

    def _W_update(self):
        """Non-negativity update"""
        self._W = np.maximum(self._rho / self._mu3 + self._image_est, 0)

    def _xi_update(self):
        # to avoid computing forward model twice
        self._xi += self._mu1 * (self._forward_out - self._X)

    def _eta_update(self):
        # to avoid finite difference operataion again?
        self._eta += self._mu2 * (self._Psi_out - self._U)

    def _rho_update(self):
        self._rho += self._mu3 * (self._image_est - self._W)

    def _update(self):
        self._U_update()
        self._X_update()
        self._image_update()

        # update forward and sparse operators
        self._forward_out = self._forward()
        self._Psi_out = self._Psi(self._image_est)

        self._W_update()
        self._xi_update()
        self._eta_update()
        self._rho_update()

    def _form_image(self):
        image = self._crop(self._image_est)
        image[image < 0] = 0
        return image.squeeze()

    def set_data(self, data):
        if not self._is_rgb:
            assert len(data.shape) == 2
            data = data[:, :, np.newaxis]
        assert len(self._psf_shape) == len(data.shape)
        self._data = self._pad(data)
        self.reset()


def soft_thresh(x, thresh):
    # numpy automatically applies functions to each element of the array
    return np.sign(x) * np.maximum(0, np.abs(x) - thresh)


def finite_diff(x):
    """Gradient of image estimate, approximated by finite difference. Space where image is assumed sparse."""
    return np.stack(
        (np.roll(x, 1, axis=0) - x, np.roll(x, 1, axis=1) - x),
        axis=len(x.shape),
    )


def finite_diff_adj(x):
    diff1 = np.roll(x[..., 0], -1, axis=0) - x[..., 0]
    diff2 = np.roll(x[..., 1], -1, axis=1) - x[..., 1]
    return diff1 + diff2


def finite_diff_gram(shape, dtype=np.float32):
    gram = np.zeros(shape, dtype=dtype)
    gram[0, 0] = 4
    gram[0, 1] = gram[1, 0] = gram[0, -1] = gram[-1, 0] = -1
    return fft.rfft2(gram, axes=(0, 1))



import cv2
import csv
import numpy as np
from lensless.constants import RPI_HQ_CAMERA_CCM_MATRIX, RPI_HQ_CAMERA_BLACK_LEVEL


SUPPORTED_BIT_DEPTH = np.array([8, 10, 12, 16])
FLOAT_DTYPES = [np.float32, np.float64]


def resize(img, factor=None, shape=None, interpolation=cv2.INTER_CUBIC):
    """
    Resize by given factor.

    Parameters
    ----------
    img :py:class:`~numpy.ndarray`
        Downsampled image.
    factor : int or float
        Resizing factor.
    shape : tuple
        (Height, width).
    interpolation : OpenCV interpolation method
        See https://docs.opencv.org/2.4/modules/imgproc/doc/geometric_transformations.html#cv2.resize

    Returns
    -------
    img :py:class:`~numpy.ndarray`
        Resized image.
    """
    min_val = img.min()
    max_val = img.max()
    img_shape = np.array(img.shape)[:2]
    if shape is None:
        assert factor is not None
        new_shape = tuple((img_shape * factor).astype(int))
        new_shape = new_shape[::-1]
        resized = cv2.resize(img, dsize=new_shape, interpolation=interpolation)
    else:
        if np.array_equal(img_shape, shape[::-1]):
            return img
        resized = cv2.resize(img, dsize=shape[::-1], interpolation=interpolation)
    return np.clip(resized, min_val, max_val)


def rgb2gray(rgb, weights=None):
    """
    Convert RGB array to grayscale.

    Parameters
    ----------
    rgb : :py:class:`~numpy.ndarray`
        (N_height, N_width, N_channel) image.
    weights : :py:class:`~numpy.ndarray`
        [Optional] (3,) weights to convert from RGB to grayscale.

    Returns
    -------
    img :py:class:`~numpy.ndarray`
        Grayscale image of dimension (height, width).

    """
    if weights is None:
        weights = np.array([0.299, 0.587, 0.144])
    assert len(weights) == 3
    return np.tensordot(rgb, weights, axes=((2,), 0))


def gamma_correction(vals, gamma=2.2):
    """
    Tutorials
    - https://www.cambridgeincolour.com/tutorials/gamma-correction.htm
    - (code, for images) https://www.pyimagesearch.com/2015/10/05/opencv-gamma-correction/
        https://lindevs.com/apply-gamma-correction-to-an-image-using-opencv/
    - (code) http://www.fourmilab.ch/documents/specrend/specrend.c

    Parameters
    ----------
    vals : array_like
        RGB values to gamma correct.
    gamma : float, optional
            Gamma correction factor to apply for plots. Default is None.

    Returns
    -------
    vals : array_like
        Gamma-corrected data.

    """

    # simple approach
    # return np.power(vals, 1 / gamma)

    # Rec. 709 gamma correction
    # http://www.fourmilab.ch/documents/specrend/specrend.c
    cc = 0.018
    inv_gam = 1 / gamma
    clip_val = (1.099 * np.power(cc, inv_gam) - 0.099) / cc
    return np.where(vals < cc, vals * clip_val, 1.099 * np.power(vals, inv_gam) - 0.099)


def get_max_val(img, nbits=None):
    """For uint image"""
    assert img.dtype not in FLOAT_DTYPES
    if nbits is None:
        nbits = int(np.ceil(np.log2(img.max())))

    if nbits not in SUPPORTED_BIT_DEPTH:
        nbits = SUPPORTED_BIT_DEPTH[nbits < SUPPORTED_BIT_DEPTH][0]
    max_val = 2**nbits - 1
    if img.max() > max_val:
        new_nbit = int(np.ceil(np.log2(img.max())))
        print(f"Detected pixel value larger than {nbits}-bit range, using {new_nbit}-bit range.")
        max_val = 2**new_nbit - 1
    return max_val


def bayer2rgb(
    img,
    nbits,
    blue_gain=None,
    red_gain=None,
    black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
    ccm=RPI_HQ_CAMERA_CCM_MATRIX,
    nbits_out=None,
):
    """
    Convert raw Bayer data to RGB with the following steps:
    - Demosaic with bi-linear interpolation, mapping the Bayer array to RGB.
    - Black level removal.
    - White balancing, applying gains to red and blue channels.
    - Color correction matrix.
    - Clip

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        2D Bayer data to convert to RGB.
    nbits : int
        Bit depth of input data.
    blue_gain : float
        Blue gain.
    red_gain : float
        Red gain.
    black_level : float
        Black level. Default is to use that of Raspberry Pi HQ camera.
    ccm : :py:class:`~numpy.ndarray`
        Color correction matrix. Default is to use that of Raspberry Pi HQ camera.
    nbits_out : int
        Output bit depth. Default is to use that of input.

    Returns
    -------
    rgb : :py:class:`~numpy.ndarray`
        RGB data.
    """
    assert len(img.shape) == 2, img.shape
    if nbits_out is None:
        nbits_out = nbits
    if nbits_out > 8:
        dtype = np.uint16
    else:
        dtype = np.uint8

    # demosaic Bayer data
    img = cv2.cvtColor(img, cv2.COLOR_BayerRG2RGB)

    # correction
    img = img - black_level
    if red_gain:
        img[:, :, 0] *= red_gain
    if blue_gain:
        img[:, :, 2] *= blue_gain
    img = img / (2**nbits - 1 - black_level)
    img[img > 1] = 1
    img = (img.reshape(-1, 3, order="F") @ ccm.T).reshape(img.shape, order="F")
    img[img < 0] = 0
    img[img > 1] = 1
    return (img * (2**nbits_out - 1)).astype(dtype)


def get_distro():
    # https://majornetwork.net/2019/11/get-linux-distribution-name-and-version-with-python/
    RELEASE_DATA = {}
    with open("/etc/os-release") as f:
        reader = csv.reader(f, delimiter="=")
        for row in reader:
            if row:
                RELEASE_DATA[row[0]] = row[1]
    if RELEASE_DATA["ID"] in ["debian", "raspbian"]:
        with open("/etc/debian_version") as f:
            DEBIAN_VERSION = f.readline().strip()
        major_version = DEBIAN_VERSION.split(".")[0]
        version_split = RELEASE_DATA["VERSION"].split(" ", maxsplit=1)
        if version_split[0] == major_version:
            # Just major version shown, replace it with the full version
            RELEASE_DATA["VERSION"] = " ".join([DEBIAN_VERSION] + version_split[1:])
    return f"{RELEASE_DATA['NAME']} {RELEASE_DATA['VERSION']}"


def print_image_info(img):
    print("dimensions : {}".format(img.shape))
    print("data type : {}".format(img.dtype))
    print("max  : {}".format(img.max()))
    print("min  : {}".format(img.min()))
    print("mean : {}".format(img.mean()))



import numpy as np

# estimated here: https://www.strollswithmydog.com/open-raspberry-pi-high-quality-camera-raw
RPI_HQ_CAMERA_CCM_MATRIX = np.array(
    [
        [2.0659, -0.93119, -0.13421],
        [-0.11615, 1.5593, -0.44314],
        [0.073694, -0.4368, 1.3636],
    ]
)
RPI_HQ_CAMERA_BLACK_LEVEL = 256.3



import numpy as np
import warnings


def autocorr2d(vals, pad_mode="reflect"):
    """
    Compute 2-D autocorrelation of image via the FFT.

    Parameters
    ----------
    vals : py:class:`~numpy.ndarray`
        2-D image.
    pad_mode : str
        Desired padding. See NumPy documentation: https://numpy.org/doc/stable/reference/generated/numpy.pad.html

    Return
    ------
    autocorr : py:class:`~numpy.ndarray`
    """
    warnings.warn("`autocorr2d` to complete!")
    return np.ones_like(vals)



from lensless.io import load_data
import numpy as np

psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"
downsample = 8


def test_load_data():
    for gray in [True, False]:
        for dtype in [np.float32, np.float64]:
            psf, data = load_data(
                psf_fp=psf_fp,
                data_fp=data_fp,
                downsample=downsample,
                plot=False,
                gray=gray,
                dtype=dtype,
            )
            if gray:
                assert len(psf.shape) == 2
            else:
                assert len(psf.shape) == 3
            assert psf.shape == data.shape
            assert psf.dtype == dtype, dtype
            assert data.dtype == dtype, dtype


test_load_data()



import abc
import numpy as np
import pathlib as plib
import matplotlib.pyplot as plt
from scipy.fftpack import next_fast_len
from lensless.plot import plot_image


class ReconstructionAlgorithm(abc.ABC):
    """
    Abstract class for defining lensless imaging reconstruction algorithms.

    The following abstract methods need to be defined:
    - `_update`: updating state variables at each iterations.
    - `reset`: reset state variables.
    - `_form_image`: any pre-processing that needs to be done in order to view
    the image estimate, e.g. reshaping or clipping.

    One advantage of deriving from this abstract class is that functionality for
    iterating, saving, and visualization is already implemented, namely in the
    `apply` method.

    Consequently, using a reconstruction algorithm that derives from it boils down
    to three steps:

    1. Creating an instance of the reconstruction algorithm.
    2. Setting the data.
    3. Applying the algorithm.

    For example, for ADMM (full example in `scripts/admm.py`):
    ```python
        recon = ADMM(psf)
        recon.set_data(data)
        res = recon.apply(n_iter=n_iter)
    ```

    A template for applying a reconstruction algorithm (including loading the
    data) can be found in `scripts/reconstruction_template.py`.

    """

    def __init__(self, psf, dtype=np.float32):
        """
        Base constructor. Derived constructor may define new state variables
        here and also reset them in `reset`.

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray`
            Point spread function (PSF) that models forward propagation.
            2D (grayscale) or 3D (RGB) data can be provided and the shape will
            be used to determine which reconstruction (and allocate the
            appropriate memory).
        dtype : float32 or float64
            Data type to use for optimization.
        """

        self._is_rgb = True if len(psf.shape) == 3 else False
        if self._is_rgb:
            self._psf = psf
            self._n_channels = 3
        else:
            self._psf = psf[:, :, np.newaxis]
            self._n_channels = 1
        self._psf_shape = np.array(self._psf.shape)

        if dtype:
            self._psf = self._psf.astype(dtype)
            self._dtype = dtype
        else:
            self._dtype = self._psf.dtype
        if self._dtype == np.float32 or dtype == "float32":
            self._complex_dtype = np.complex64
        elif self._dtype == np.float64 or dtype == "float64":
            self._complex_dtype = np.complex128
        else:
            raise ValueError(f"Unsupported dtype : {self._dtype}")

        # cropping / padding indices
        self._padded_shape = 2 * self._psf_shape[:2] - 1
        self._padded_shape = np.array([next_fast_len(i) for i in self._padded_shape])
        self._padded_shape = np.r_[self._padded_shape, [self._n_channels]]
        self._start_idx = (self._padded_shape[:2] - self._psf_shape[:2]) // 2
        self._end_idx = self._start_idx + self._psf_shape[:2]

        # pre-compute operators / outputs
        self._image_est = None
        self._data = None
        self.reset()

    @abc.abstractmethod
    def reset(self):
        """Reset state variables."""
        return

    @abc.abstractmethod
    def _update(self):
        """Update state variables."""
        return

    @abc.abstractmethod
    def _form_image(self):
        """
        Any pre-processing to form a viewable image, e.g. reshaping or clipping.
        """
        return

    def set_data(self, data):
        """
        Set lensless data for recontruction.

        Parameters
        ----------
        data : :py:class:`~numpy.ndarray`
            Lensless data on which to iterate to recover an estimate of the
             scene. Should match provide PSF, i.e. shape and 2D (grayscale) or
             3D (RGB).
        """
        if not self._is_rgb:
            assert len(data.shape) == 2
            data = data[:, :, np.newaxis]
        assert len(self._psf_shape) == len(data.shape)
        self._data = data

    def get_image_est(self):
        """Get current image estimate."""
        return self._form_image()

    def _progress(self):
        """
        Optional method for printing progress update, e.g. relative improvement
        in reconstruction.
        """
        return

    def apply(
        self, n_iter=100, disp_iter=10, plot_pause=0.2, plot=True, save=False, gamma=None, ax=None
    ):
        """
        Method for performing iterative reconstruction. Note that `set_data`
        must be called beforehand.

        Parameters
        ----------
        n_iter : int
            Number of iterations.
        disp_iter : int
            How often to display and/or intermediate reconstruction (in number
            of iterations). If `None` OR `plot` or `save` are False, no
            intermediate reconstruction will be plotted/saved.
        plot_pause : float
            Number of seconds to pause after displaying reconstruction.
        plot : bool
            Whether to plot final result, and intermediate results if
            `disp_iter` is not None.
        save : bool
            Whether to save final result (as PNG), and intermediate results if
            `disp_iter` is not None.
        gamma : float, optional
            Gamma correction factor to apply for plots. Default is None.
        ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill for plotting/saving, default is to create one.

        Returns
        -------
        final_im : :py:class:`~numpy.ndarray`
            Final reconstruction.
        ax : :py:class:`~matplotlib.axes.Axes`
            `Axes` object on which final reconstruction is displayed. Only
            returning if `plot` or `save` is True.

        """
        assert self._data is not None, "Must set data with `set_data()`"

        if (plot or save) and disp_iter is not None:
            if ax is None:
                ax = plot_image(self._data, gamma=gamma)
        else:
            ax = None
            disp_iter = n_iter + 1

        for i in range(n_iter):
            self._update()

            if (plot or save) and (i + 1) % disp_iter == 0:
                self._progress()
                img = self._form_image()
                ax = plot_image(img, ax=ax, gamma=gamma)
                ax.set_title("Reconstruction after iteration {}".format(i + 1))
                if save:
                    plt.savefig(plib.Path(save) / f"{i + 1}.png")
                if plot:
                    plt.draw()
                    plt.pause(plot_pause)

        final_im = self._form_image()
        if plot or save:
            ax = plot_image(final_im, ax=ax, gamma=gamma)
            ax.set_title("Final reconstruction after {} iterations".format(n_iter))
            if save:
                plt.savefig(plib.Path(save) / f"{n_iter}.png")
            return final_im, ax
        else:
            return final_im



import numpy as np
from lensless.recon import ReconstructionAlgorithm
import inspect
from scipy import fft


class GradientDescentUpdate:
    """Gradient descent update techniques."""

    VANILLA = "vanilla"
    NESTEROV = "nesterov"
    FISTA = "fista"

    @staticmethod
    def all_values():
        vals = []
        for i in inspect.getmembers(GradientDescentUpdate):
            # remove private and protected functions, and this function
            if not i[0].startswith("_") and not callable(i[1]):
                vals.append(i[1])
        return vals


def non_neg(xi):
    """
    Clip input so that it is non-negative.

    Parameters
    ----------
    xi : :py:class:`~numpy.ndarray`
        Data to clip.

    Returns
    -------
    nonneg : :py:class:`~numpy.ndarray`
        Non-negative projection of input.

    """
    xi = np.maximum(xi, 0)
    return xi


class GradientDescient(ReconstructionAlgorithm):
    def __init__(self, psf, dtype=np.float32, proj=non_neg, **kwargs):
        """
        Object for applying projected gradient descent.

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray`
            Point spread function (PSF) that models forward propagation.
            2D (grayscale) or 3D (RGB) data can be provided and the shape will
            be used to determine which reconstruction (and allocate the
            appropriate memory).
        dtype : float32 or float64
            Data type to use for optimization.
        proj : :py:class:`function`
            Projection function to apply at each iteration. Default is
            non-negative.
        """

        super(GradientDescient, self).__init__(psf, dtype)
        assert callable(proj)
        self._proj = proj

    def _crop(self, x):
        return x[self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]]

    def _pad(self, v):
        vpad = np.zeros(self._padded_shape).astype(v.dtype)
        vpad[self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1]] = v
        return vpad

    def reset(self):

        # initial guess, half intensity image
        # for online approach could use last reconstruction
        psf_flat = self._psf.reshape(-1, self._n_channels)
        pixel_start = (np.max(psf_flat, axis=0) + np.min(psf_flat, axis=0)) / 2
        x = np.ones(self._psf_shape, dtype=self._dtype) * pixel_start
        self._image_est = self._pad(x)

        # spatial frequency response
        self._H = fft.rfft2(
            self._pad(self._psf), norm="ortho", axes=(0, 1), s=self._padded_shape[:2]
        )
        self._Hadj = np.conj(self._H)

        Hadj_flat = self._Hadj.reshape(-1, self._n_channels)
        H_flat = self._H.reshape(-1, self._n_channels)
        self._alpha = np.real(1.8 / np.max(Hadj_flat * H_flat, axis=0))

    def _grad(self):
        diff = self._forward() - self._data
        return self._backward(diff)

    def _forward(self):
        Vk = fft.rfft2(self._image_est, axes=(0, 1), s=self._padded_shape[:2])
        return self._crop(
            fft.ifftshift(
                fft.irfft2(self._H * Vk, axes=(0, 1), s=self._padded_shape[:2]), axes=(0, 1)
            )
        )

    def _backward(self, x):
        X = fft.rfft2(self._pad(x), axes=(0, 1), s=self._padded_shape[:2])
        return fft.ifftshift(
            fft.irfft2(self._Hadj * X, axes=(0, 1), s=self._padded_shape[:2]), axes=(0, 1)
        )

    def _update(self):
        self._image_est -= self._alpha * self._grad()
        self._image_est = self._proj(self._image_est)

    def _form_image(self):
        return self._proj(self._crop(self._image_est)).squeeze()


class NesterovGradientDescent(GradientDescient):
    """
    Object for applying projected gradient descent with Nesterov momentum for
    acceleration.

    Tutorial on Nesterov momentum: https://machinelearningmastery.com/gradient-descent-with-nesterov-momentum-from-scratch/

    """

    def __init__(self, psf, dtype=np.float32, proj=non_neg, p=0, mu=0.9, **kwargs):
        self._p = p
        self._mu = mu
        super(NesterovGradientDescent, self).__init__(psf, dtype, proj)

    def reset(self, p=0, mu=0.9):
        self._p = p
        self._mu = mu
        super(NesterovGradientDescent, self).reset()

    def _update(self):
        p_prev = self._p
        self._p = self._mu * self._p - self._alpha * self._grad()
        self._image_est += -self._mu * p_prev + (1 + self._mu) * self._p
        self._image_est = self._proj(self._image_est)


class FISTA(GradientDescient):
    """
    Object for applying projected gradient descent with FISTA (Fast Iterative
    Shrinkage-Thresholding Algorithm) for acceleration.

    Paper: https://www.ceremade.dauphine.fr/~carlier/FISTA

    """

    def __init__(self, psf, dtype=np.float32, proj=non_neg, tk=1, **kwargs):

        super(FISTA, self).__init__(psf, dtype, proj)
        self._tk = tk
        self._xk = self._image_est

    def reset(self, tk=1):
        super(FISTA, self).reset()
        self._tk = tk
        self._xk = self._image_est

    def _update(self):
        self._image_est -= self._alpha * self._grad()
        xk = self._proj(self._image_est)
        tk = (1 + np.sqrt(1 + 4 * self._tk**2)) / 2
        self._image_est = xk + (self._tk - 1) / tk * (xk - self._xk)
        self._tk = tk
        self._xk = xk



from .recon import ReconstructionAlgorithm
from .admm import ADMM
from .gradient_descent import (
    GradientDescient,
    NesterovGradientDescent,
    FISTA,
    GradientDescentUpdate,
)

try:
    from .apgd import APGD, APGDPriors

    pycsou_available = True
except:
    pycsou_available = False



from lensless.recon import ReconstructionAlgorithm
import inspect
import numpy as np
from pycsou.linop.conv import Convolve2D
from pycsou.func.loss import SquaredL2Loss
from pycsou.func.penalty import NonNegativeOrthant, SquaredL2Norm, L1Norm
from pycsou.opt.proxalgs import APGD as APGD_pyc
from copy import deepcopy

try:
    from lensless.realfftconv_sol import RealFFTConvolve2D
except:
    from lensless.realfftconv import RealFFTConvolve2D


class APGDPriors:
    """
    Priors (compatible with Pycsou) for APGD.

    See Pycsou documentation for available penalties:
    https://matthieumeo.github.io/pycsou/html/api/functionals/pycsou.func.penalty.html?highlight=penalty#module-pycsou.func.penalty
    """

    L2 = "l2"
    NONNEG = "nonneg"
    L1 = "l1"

    @staticmethod
    def all_values():
        vals = []
        for i in inspect.getmembers(APGDPriors):
            # remove private and protected functions, and this function
            if not i[0].startswith("_") and not callable(i[1]):
                vals.append(i[1])
        return vals


class APGD(ReconstructionAlgorithm):
    def __init__(
        self,
        psf,
        max_iter=500,
        dtype=np.float32,
        diff_penalty=None,
        prox_penalty=APGDPriors.NONNEG,
        acceleration="BT",
        diff_lambda=0.001,
        prox_lambda=0.001,
        realconv=True,
        **kwargs
    ):
        """
        Wrapper for Pycsou's APGD (accelerated proximal gradient descent)
        applied to lensless imaging.

        Pycsou APGD documentation: https://matthieumeo.github.io/pycsou/html/api/algorithms/pycsou.opt.proxalgs.html?highlight=apgd#pycsou.opt.proxalgs.AcceleratedProximalGradientDescent

        Parameters
        ----------
        psf :py:class:`~numpy.ndarray`
            PSF that models forward propagation.
        max_iter : int, optional
            Maximal number of iterations.
        dtype : float32 or float64
            Data type to use for optimization.
        diff_penalty : None or str or :py:class:`~pycsou.func.penalty`
            Differentiable functional to serve as prior / regularization term.
            Default is None. See Pycsou documentation for available
            penalties: https://matthieumeo.github.io/pycsou/html/api/functionals/pycsou.func.penalty.html?highlight=penalty#module-pycsou.func.penalty
        prox_penalty : None or str or :py:class:`~pycsou.func.penalty`
            Proximal functional to serve as prior / regularization term. Default
            is non-negative prior. See Pycsou documentation for available
            penalties: https://matthieumeo.github.io/pycsou/html/api/functionals/pycsou.func.penalty.html?highlight=penalty#module-pycsou.func.penalty
        acceleration : [None, 'BT', 'CD']
            Which acceleration scheme should be used (None for no acceleration).
            "BT" (Beck and Teboule) has convergence `O(1/k^2)`, while "CD"
            (Chambolle and Dossal) has convergence `o(1/K^2)`. So "CD" should be
            faster. but from our experience "BT" gives better results.
        diff_lambda : float
            Weight of differentiable penalty.
        prox_lambda : float
            Weight of proximal penalty.
        realconv : bool
            Whether to apply convolution for real signals (if available).
        """

        # PSF and data are the same size / shape
        self._original_shape = psf.shape
        self._original_size = psf.size

        self._apgd = None
        self._gen = None

        super(APGD, self).__init__(psf, dtype)

        self._max_iter = max_iter

        # Convolution operator
        if realconv:
            self._H = RealFFTConvolve2D(self._psf, dtype=dtype)
        else:
            assert self._is_rgb is False, "RGB not supported for `Convolve2D`."
            self._H = Convolve2D(size=psf.size, filter=psf, shape=psf.shape, dtype=dtype)
        self._H.compute_lipschitz_cst()

        # initialize solvers which will be created when data is set
        if diff_penalty is not None:
            if diff_penalty == APGDPriors.L2:
                self._diff_penalty = diff_lambda * SquaredL2Norm(dim=self._H.shape[1])
            else:
                assert hasattr(diff_penalty, "jacobianT")
                self._diff_penalty = diff_lambda * diff_penalty(dim=self._H.shape[1])
        else:
            self._diff_penalty = None

        if prox_penalty is not None:
            if prox_penalty == APGDPriors.L1:
                self._prox_penalty = prox_lambda * L1Norm(dim=self._H.shape[1])
            elif prox_penalty == APGDPriors.NONNEG:
                self._prox_penalty = prox_lambda * NonNegativeOrthant(dim=self._H.shape[1])
            else:
                try:
                    self._prox_penalty = prox_lambda * prox_penalty(dim=self._H.shape[1])
                except:
                    raise ValueError("Unexpected prior.")
        else:
            self._prox_penalty = None

        self._acc = acceleration

    def set_data(self, data):
        """
        For `APGD`, we use data to initialize problem for Pycsou.

        Parameters
        ----------
        data : :py:class:`~numpy.ndarray`
            Lensless data on which to iterate to recover an estimate of the
             scene. Should match provide PSF, i.e. shape and 2D (grayscale) or
             3D (RGB).

        """
        if not self._is_rgb:
            assert len(data.shape) == 2
            data = data[:, :, np.newaxis]
        assert len(self._psf_shape) == len(data.shape)
        self._data = data

        """ Set up problem """
        # Cost function
        loss = (1 / 2) * SquaredL2Loss(dim=self._H.shape[0], data=self._data.ravel())
        F = loss * self._H
        if self._diff_penalty is not None:
            F += self._diff_penalty

        if self._prox_penalty is not None:
            G = self._prox_penalty
            dim = G.shape[1]
        else:
            G = None
            dim = self._data.size

        self._apgd = APGD_pyc(dim=dim, F=F, G=G, acceleration=self._acc)

        # -- setup to print progress report
        self._apgd.old_iterand = deepcopy(self._apgd.init_iterand)
        self._apgd.update_diagnostics()
        self._gen = self._apgd.iterates(n=self._max_iter)

    def reset(self):
        self._image_est = np.zeros(self._original_size, dtype=self._dtype)
        if self._apgd is not None:
            self._apgd.reset()

            # -- setup to print progress report
            self._apgd.old_iterand = deepcopy(self._apgd.init_iterand)
            self._apgd.update_diagnostics()
            self._gen = self._apgd.iterates(n=self._max_iter)

    def _progress(self):
        """
        Pycsou has functionality for printing progress that we will make use of
        here.

        """
        self._apgd.update_diagnostics()
        self._apgd.old_iterand = deepcopy(self._apgd.iterand)
        self._apgd.print_diagnostics()

    def _update(self):
        next(self._gen)
        self._image_est[:] = self._apgd.iterand["iterand"]

    def _form_image(self):
        image = self._image_est.reshape(self._original_shape)
        image[image < 0] = 0
        return image



import os.path
import rawpy
import cv2
import numpy as np
from lensless.util import resize, bayer2rgb, rgb2gray, print_image_info
from lensless.plot import plot_image
from lensless.constants import RPI_HQ_CAMERA_CCM_MATRIX, RPI_HQ_CAMERA_BLACK_LEVEL


def load_image(
    fp,
    verbose=False,
    flip=False,
    bayer=False,
    black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
    blue_gain=None,
    red_gain=None,
    ccm=RPI_HQ_CAMERA_CCM_MATRIX,
    back=None,
    nbits_out=None,
):
    """
    Load image as numpy array.

    Parameters
    ----------
    fp : str
        Full path to file.
    verbose : bool, optional
        Whether to plot into about file.
    flip : bool
        Whether to flip data (vertical and horizontal).
    bayer : bool
        Whether input data is Bayer.
    blue_gain : float
        Blue gain for color correction.
    red_gain : float
        Red gain for color correction.
    black_level : float
        Black level. Default is to use that of Raspberry Pi HQ camera.
    ccm : :py:class:`~numpy.ndarray`
        Color correction matrix. Default is to use that of Raspberry Pi HQ camera.
    back : array_like
        Background level to subtract.
    nbits_out : int
        Output bit depth. Default is to use that of input.

    Returns
    -------
    img :py:class:`~numpy.ndarray`
        RGB image of dimension (height, width, 3).
    """
    assert os.path.isfile(fp)
    if "dng" in fp:
        assert bayer
        raw = rawpy.imread(fp)
        img = raw.raw_image
        # TODO : use raw.postprocess?
        ccm = raw.color_matrix[:, :3]
        black_level = np.array(raw.black_level_per_channel[:3]).astype(np.float32)
    else:
        img = cv2.imread(fp, cv2.IMREAD_UNCHANGED)

    if bayer:
        assert len(img.shape) == 2, img.shape
        if img.max() > 255:
            # HQ camera
            n_bits = 12
        else:
            n_bits = 8

        if back:
            back_img = cv2.imread(back, cv2.IMREAD_UNCHANGED)
            dtype = img.dtype
            img = img.astype(np.float32) - back_img.astype(np.float32)
            img = np.clip(img, a_min=0, a_max=img.max())
            img = img.astype(dtype)
        if nbits_out is None:
            nbits_out = n_bits
        img = bayer2rgb(
            img,
            nbits=n_bits,
            blue_gain=blue_gain,
            red_gain=red_gain,
            black_level=black_level,
            ccm=ccm,
            nbits_out=nbits_out,
        )

    else:
        if len(img.shape) == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    if flip:
        img = np.flipud(img)
        img = np.fliplr(img)

    if verbose:
        print_image_info(img)

    return img


def load_psf(
    fp,
    downsample=1,
    return_float=True,
    bg_pix=(5, 25),
    return_bg=False,
    flip=False,
    verbose=False,
    bayer=False,
    blue_gain=None,
    red_gain=None,
    dtype=np.float32,
    nbits_out=None,
    single_psf=False,
    shape=None,
):
    """
    Load and process PSF for analysis or for reconstruction.

    Basic steps are:
    - Load image.
    - (Optionally) subtract background. Recommended.
    - (Optionally) resize to more manageable size
    - (Optionally) normalize within [0, 1] if using for reconstruction; otherwise cast back to uint for analysis.

    Parameters
    ----------
    fp : str
        Full path to file.
    downsample : int, optional
        Downsampling factor. Recommended for image reconstruction.
    return_float : bool, optional
        Whether to return PSF as float array, or unsigned int.
    bg_pix : tuple, optional
        Section of pixels to take from top left corner to remove background level. Set to `None` to omit this
        step, althrough it is highly recommended.
    return_bg : bool, optional
        Whether to return background level, for removing from data for reconstruction.
    flip : bool, optional
        Whether to flip up-down and left-right.
    verbose : bool
        Whether to print metadata.
    bayer : bool
        Whether input data is Bayer.
    blue_gain : float
        Blue gain for color correction.
    red_gain : float
        Red gain for color correction.
    dtype : float32 or float64
        Data type of returned data.
    nbits_out : int
        Output bit depth. Default is to use that of input.
    single_psf : bool
        Whether to sum RGB channels into single PSF, same across channels. Done
        in "Learned reconstructions for practical mask-based lensless imaging"
        of Kristina Monakhova et. al.

    Returns
    -------
    psf :py:class:`~numpy.ndarray`
        2-D array of PSF.
    """

    # load image data and extract necessary channels
    psf = load_image(
        fp,
        verbose=verbose,
        flip=flip,
        bayer=bayer,
        blue_gain=blue_gain,
        red_gain=red_gain,
        nbits_out=nbits_out,
    )

    original_dtype = psf.dtype
    psf = np.array(psf, dtype=dtype)

    # subtract background, assume black edges
    bg = np.zeros(3)
    if bg_pix is not None:
        bg = []
        for i in range(3):
            bg_i = np.mean(psf[bg_pix[0] : bg_pix[1], bg_pix[0] : bg_pix[1], i])
            psf[:, :, i] -= bg_i
            bg.append(bg_i)
        psf = np.clip(psf, a_min=0, a_max=psf.max())
        bg = np.array(bg)

    # resize
    if shape:
        psf = resize(psf, shape=shape)
    elif downsample != 1:
        psf = resize(psf, factor=1 / downsample)

    if single_psf:
        assert len(psf.shape) == 3
        # TODO : in Lensless Learning, they sum channels --> `psf_diffuser = np.sum(psf_diffuser,2)`
        # https://github.com/Waller-Lab/LenslessLearning/blob/master/pre-trained%20reconstructions.ipynb
        psf = np.sum(psf, 2)
        psf = psf[:, :, np.newaxis]

    # normalize
    if return_float:
        # psf /= psf.max()
        psf /= np.linalg.norm(psf.ravel())
    else:
        psf = psf.astype(original_dtype)

    if return_bg:
        return psf, bg
    else:
        return psf


def load_data(
    psf_fp,
    data_fp,
    downsample=None,
    bg_pix=(5, 25),
    plot=True,
    flip=False,
    bayer=False,
    blue_gain=None,
    red_gain=None,
    gamma=None,
    gray=False,
    dtype=np.float32,
    single_psf=False,
    shape=None,
):
    """
    Load data for image reconstruction.

    Parameters
    ----------
    psf_fp : str
        Full path to PSF file.
    data_fp : str
        Full path to measurement file.
    downsample : int or float
        Downsampling factor.
    bg_pix : tuple, optional
        Section of pixels to take from top left corner to remove background
        level. Set to `None` to omit this step, although it is highly
        recommended.
    plot : bool, optional
        Whether or not to plot PSF and raw data.
    flip : bool
        Whether to flip data (vertical and horizontal).
    bayer : bool
        Whether input data is Bayer.
    blue_gain : float
        Blue gain for color correction.
    red_gain : float
        Red gain for color correction.
    gamma : float, optional
        Optional gamma factor to apply, ONLY for plotting. Default is None.
    gray : bool
        Whether to load as grayscale or RGB.
    dtype : float32 or float64
        Data type of returned data.
    single_psf : bool
        Whether to sum RGB channels into single PSF, same across channels. Done
        in "Learned reconstructions for practical mask-based lensless imaging"
        of Kristina Monakhova et. al.

    Returns
    -------
    psf :py:class:`~numpy.ndarray`
        2-D array of PSF.
    data :py:class:`~numpy.ndarray`
        2-D array of raw measurement data.
    """

    assert os.path.isfile(psf_fp)
    assert os.path.isfile(data_fp)
    if shape is None:
        assert downsample is not None

    # load and process PSF data
    psf, bg = load_psf(
        psf_fp,
        downsample=downsample,
        return_float=True,
        bg_pix=bg_pix,
        return_bg=True,
        flip=flip,
        bayer=bayer,
        blue_gain=blue_gain,
        red_gain=red_gain,
        dtype=dtype,
        single_psf=single_psf,
        shape=shape,
    )

    # load and process raw measurement
    data = load_image(data_fp, flip=flip, bayer=bayer, blue_gain=blue_gain, red_gain=red_gain)
    data = np.array(data, dtype=dtype)

    data -= bg
    data = np.clip(data, a_min=0, a_max=data.max())
    if data.shape != psf.shape:
        # in DiffuserCam dataset, images are already reshaped
        data = resize(data, shape=psf.shape[:2])
    data /= np.linalg.norm(data.ravel())

    if gray:
        psf = rgb2gray(psf)
        data = rgb2gray(data)

    if plot:
        ax = plot_image(psf, gamma=gamma)
        ax.set_title("PSF")
        ax = plot_image(data, gamma=gamma)
        ax.set_title("Raw data")

    psf = np.array(psf, dtype=dtype)
    data = np.array(data, dtype=dtype)

    return psf, data



import numpy as np
from pycsou.core.linop import LinearOperator
from typing import Union, Optional
from numbers import Number


class RealFFTConvolve2D(LinearOperator):
    def __init__(self, filter, dtype: Optional[type] = None):
        """
        Linear operator that performs convolution in Fourier domain, and assumes
        real-valued signals.

        Parameters
        ----------
        filter :py:class:`~numpy.ndarray`
            2D filter to use. Must be of shape (height, width, channels) even if
            only one channel.
        dtype : float32 or float64
            Data type to use for optimization.
        """

        assert len(filter.shape) == 3

        # precompute filter in frequency domain
        raise NotImplementedError

        # call parent class
        super(RealFFTConvolve2D, self).__init__(shape=shape, dtype=dtype)

    def __call__(self, x: Union[Number, np.ndarray]) -> Union[Number, np.ndarray]:
        # like here: https://github.com/PyLops/pylops/blob/3e7eb22a62ec60e868ccdd03bc4b54806851cb26/pylops/signalprocessing/ConvolveND.py#L103
        raise NotImplementedError

    def adjoint(self, y: Union[Number, np.ndarray]) -> Union[Number, np.ndarray]:
        raise NotImplementedError



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lensless",
    version="1.0.2",
    author="Eric Bezzam",
    author_email="ebezzam@gmail.com",
    description="Package to control and image with a lensless camera running on a Raspberry Pi.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LCAV/LenslessPiCam",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "opencv-python==4.5.1.48",
        "numpy>=1.21",
        "scipy==1.7.0",
        "click==8.0.1",
        "image==1.5.33",
        "picamerax==20.9.1",
        "matplotlib==3.4.2",
    ],
    extra_requires={"dev": ["pudb", "black"]},
)



import numpy as np
from skimage.metrics import mean_squared_error, peak_signal_noise_ratio, structural_similarity
import lpips as lpips_lib
import torch
from scipy.ndimage import rotate
from lensless.util import resize


def mse(true, est, normalize=True):
    """
    Compute the mean-squared error between two images. The closer to 0, the
    closer the match.

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    mse : float
        The mean-squared error (MSE) metric.

    """
    if normalize:
        true /= true.max()
        est /= est.max()
    return mean_squared_error(image0=true, image1=est)


def psnr(true, est, normalize=True):
    """
    Compute the peak signal to noise ratio (PSNR) for an image. The higher the
    value, the better the match.

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    psnr : float
        The PSNR metric.

    """
    if normalize:
        true /= true.max()
        est /= est.max()
    return peak_signal_noise_ratio(image_true=true, image_test=est)


def ssim(true, est, normalize=True):
    """
    Compute the mean structural similarity index between two images. Values lie
    within [0, 1]. The closer to 1, the closer the match.

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    mssim : float
        The mean structural similarity index over the image.

    """
    if normalize:
        true /= true.max()
        est /= est.max()
    return structural_similarity(im1=true, im2=est, channel_axis=2)


def lpips(true, est, normalize=True):
    """
    Compute a perceptual metric (LPIPS) between two images. Values lie within
    [0, 1]. The closer to 0, the closer the match.

    GitHub: https://github.com/richzhang/PerceptualSimilarity

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    mssim : float
        The LPIPS metric.

    """
    if normalize:
        true /= true.max()
        est /= est.max()
    loss_fn = lpips_lib.LPIPS(net="alex", verbose=False)
    true = torch.from_numpy(
        np.transpose(true, axes=(2, 0, 1))[
            np.newaxis,
        ].copy()
    )
    est = torch.from_numpy(
        np.transpose(est, axes=(2, 0, 1))[
            np.newaxis,
        ].copy()
    )
    return loss_fn.forward(true, est).squeeze().item()


def extract(estimate, original, vertical_crop, horizontal_crop, rotation, verbose=False):
    """
    Utility function to extract matching region in reconstruction and in original
    image. Later will also be resized to the same dimensions as the estimate.

    Parameters
    ----------
    estimate : :py:class:`~numpy.ndarray`
        Reconstructed image from lensless data.
    original :py:class:`~numpy.ndarray`
        Original image.
    vertical_crop : (int, int)
        Vertical region (in pixels) to keep.
    horizontal_crop : (int, int)
        Horizontal region (in pixels) to keep.
    rotation : float
        Degrees to rotate reconstruction.
    verbose : bool

    Returns
    -------
    estimate : :py:class:`~numpy.ndarray`
        Cropped and rotated image estimate.
    img_resize : :py:class:`~numpy.ndarray`
        Original image resized that the dimensions of `estimate`.

    """

    # crop and rotate estimate image
    estimate = rotate(
        estimate[vertical_crop[0] : vertical_crop[1], horizontal_crop[0] : horizontal_crop[1]],
        angle=rotation,
    )
    estimate /= estimate.max()
    estimate = np.clip(estimate, 0, 1)
    if verbose:
        print("estimate cropped: ")
        print(estimate.shape)
        print(estimate.dtype)
        print(estimate.max())

    # resize original image accordingly
    factor = estimate.shape[1] / original.shape[1]
    if verbose:
        print("resize factor", factor)
    img_resize = np.zeros_like(estimate)
    tmp = resize(original, factor=factor).astype(estimate.dtype)
    img_resize[
        : min(estimate.shape[0], tmp.shape[0]), : min(estimate.shape[1], tmp.shape[1])
    ] = tmp[: min(estimate.shape[0], tmp.shape[0]), : min(estimate.shape[1], tmp.shape[1])]
    if verbose:
        print("\noriginal resized: ")
        print(img_resize.shape)
        print(img_resize.dtype)
        print(img_resize.max())

    return estimate, img_resize
